/*	$NetBSD: if_ed.c,v 1.100 1996/05/12 23:52:19 mycroft Exp $	*/

/*
 * Device driver for National Semiconductor DS8390/WD83C690 based ethernet
 * adapters.
 *
 * Copyright (c) 1994, 1995 Charles M. Hannum.  All rights reserved.
 *
 * Copyright (C) 1993, David Greenman.  This software may be used, modified,
 * copied, distributed, and sold, in both source and binary form provided that
 * the above copyright and these terms are retained.  Under no circumstances is
 * the author responsible for the proper functioning of this software, nor does
 * the author assume any responsibility for damages incurred with its use.
 *
 * Currently supports the Western Digital/SMC 8003 and 8013 series, the SMC
 * Elite Ultra (8216), the 3Com 3c503, the NE1000 and NE2000, and a variety of
 * similar clones.
 */

/*
 * ed_softc: per line info and status
 */
struct ed_softc {
	struct	device sc_dev;
	void *sc_ih;
	bus_chipset_tag_t sc_bc;  /* bus identifier */

	struct	arpcom sc_arpcom;	/* ethernet common */
	void *sc_sh;			/* shutdown hook */

	char	*type_str;	/* pointer to type string */
	u_char	vendor;		/* interface vendor */
	u_char	type;		/* interface type code */
	u_short	spec_flags;
#define ED_REATTACH	0x0001	/* Reattach */
#define ED_NOTPRESENT 	0x0002	/* card not present; do not allow
				   reconfiguration */

	bus_io_handle_t sc_ioh;   /* io handle */
	bus_io_handle_t sc_delayioh; /* io handle for `delay port' */
	bus_mem_handle_t sc_memh; /* bus memory handle */

	bus_io_size_t	asic_base;	/* offset of ASIC I/O port */
	bus_io_size_t	nic_base;	/* offset of NIC (DS8390) I/O port */

/*
 * The following 'proto' variable is part of a work-around for 8013EBT asics
 * being write-only.  It's sort of a prototype/shadow of the real thing.
 */
	u_char	wd_laar_proto;
/*
 * This `proto' variable is so we can turn MENB on and off without reading
 * the value back from the card all the time.
 */
	u_char	wd_msr_proto;
	u_char	cr_proto;	/* values always set in CR */
	u_char	isa16bit;	/* width of access to card 0=8 or 1=16 */
	u_char	is790;		/* set by probe if NIC is a 790 */

	int	mem_start;	/* offset of NIC memory */
	int	mem_end;	/* offset of NIC memory end */
	int	mem_size;	/* total NIC memory size */
	int	mem_ring;	/* offset of RX ring-buffer (in NIC mem) */

	u_char	mem_shared;	/* NIC memory is shared with host */
	u_char	txb_cnt;	/* number of transmit buffers */
	u_char	txb_inuse;	/* number of transmit buffers active */

	u_char 	txb_new;	/* pointer to where new buffer will be added */
	u_char	txb_next_tx;	/* pointer to next buffer ready to xmit */
	u_short	txb_len[8];	/* buffered xmit buffer lengths */
	u_char	tx_page_start;	/* first page of TX buffer area */
	u_char	rec_page_start;	/* first page of RX ring-buffer */
	u_char	rec_page_stop;	/* last page of RX ring-buffer */
	u_char	next_packet;	/* pointer to next unread RX packet */
};

extern int ed_find __P((struct ed_softc *,
			struct cfdata *,
			struct isa_attach_args *ia));
extern void edattach __P((struct ed_softc *,
			  bus_chipset_tag_t,
			  bus_io_handle_t));
extern int ed_find_Novell __P((struct ed_softc *,
			       struct cfdata *,
			       struct isa_attach_args *ia));
extern int edintr __P((void *));
