/*
 * Copyright (c) 1996 John T. Kohl.  All rights reserved.
 * Copyright (c) 1994 Stefan Grefen.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Charles Hannum.
 *	This product includes software developed by Stefan Grefen.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/*
 * COM driver, based on HP dca driver
 * uses National Semiconductor NS16450/NS16550AF UART
 */
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <sys/tty.h>
#include <sys/proc.h>
#include <sys/user.h>
#include <sys/conf.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/kernel.h>
#include <sys/syslog.h>
#include <sys/types.h>
#include <sys/device.h>

#include <machine/bus.h>
#include <machine/intr.h>

#include <dev/isa/isavar.h>
#include <dev/isa/comreg.h>
#include <dev/isa/comvar.h>

int com_isa_probe __P((struct device *, void *, void *));
void com_isa_attach __P((struct device *, struct device *, void *));

struct cfattach com_isa_ca = {
	sizeof(struct com_softc), com_isa_probe, com_isa_attach
};

/*
 * These attach routines are used for both direct ISA and PCMCIA-mediated
 * attachment. 
 */

int
com_isa_probe(parent, match, aux)
	struct device *parent;
	void *match, *aux;
{
	bus_chipset_tag_t bc;
	bus_io_handle_t ioh;
	int iobase;
	int rv = 1;

	struct isa_attach_args *ia = aux;

	bc = ia->ia_bc;
	iobase = ia->ia_iobase;

	/* if it's in use as console, it's there. */
	if (iobase != comconsaddr || comconsattached) {
		if (bus_io_map(bc, iobase, COM_NPORTS, &ioh)) {
			return 0;
		}
		rv = comprobe1(bc, ioh, iobase);
		bus_io_unmap(bc, ioh, COM_NPORTS);
	}
	if (rv) {
		ia->ia_iosize = COM_NPORTS;
		ia->ia_msize = 0;
	}
	return (rv);
}

struct com_isa_cbargs {
	isa_chipset_tag_t ia_ic;
	int irq;
};

void
com_isa_attach_callback(sc, arg)
	struct com_softc *sc;
	void *arg;
{
	struct com_isa_cbargs *cba = arg;

	if (cba->irq != IRQUNK) {
		sc->sc_ih = isa_intr_establish(cba->ia_ic, cba->irq,
					       IST_EDGE, IPL_TTY, comintr, sc);
	} else
		panic("comattach: IRQ but can't have one");
}

void
com_isa_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct com_softc *sc = (void *)self;
	int iobase;
	bus_chipset_tag_t bc;
	struct isa_attach_args *ia = aux;
	struct com_isa_cbargs cba;

	/*
	 * We're living on an isa(-type) device.  Map the ports.
	 */
	iobase = sc->sc_iobase = ia->ia_iobase;
	bc = sc->sc_bc = ia->ia_bc;

	if (iobase != comconsaddr) {
		if (bus_io_map(bc, iobase, COM_NPORTS, &sc->sc_ioh))
			panic("comattach: io mapping failed");
	} else
		sc->sc_ioh = comconsioh;

	cba.irq = ia->ia_irq;
	cba.ia_ic = ia->ia_ic;

	com_attach_subr(sc, com_isa_attach_callback, (void *)&cba);
}
