/*	$NetBSD$	*/

/*
 * Copyright (c) 1996 ENAMI, Tsugutomo (enami@ba2.so-net.or.jp)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Herb Peyerl.
 * 4. The name of Herb Peyerl may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <sys/syslog.h>
#include <sys/select.h>
#include <sys/device.h>

#include <machine/cpu.h>
#include <machine/pio.h>
#include <machine/bus.h>

#include <scsi/scsi_all.h>
#include <scsi/scsi_message.h>
#include <scsi/scsiconf.h>

#include <dev/isa/isavar.h>
#include <dev/pcmcia/pcmciavar.h>
#include <dev/ic/aic6360reg.h>
#include <dev/ic/aic6360var.h>

#if defined(DEBUG) || defined(AIC_DEBUG)
#define STATIC
#else
#define STATIC static
#endif

int aic_pcmcia_match	__P((struct device *, void *, void *));
#if 0
void aic_pcmcia_attach	__P((struct device *, struct device *, void *));
#endif
int aic_pcmcia_detach	__P((struct device *));

STATIC int aic_pcmcia_isa_attach __P((struct device *, void *,
				      void *, struct pcmcia_link *));
STATIC int aic_pcmcia_mod __P((struct pcmcia_link *, struct device *,
			       struct pcmcia_conf *, struct cfdata * cf));
STATIC int aic_remove	__P((struct pcmcia_link *, struct device *));

struct cfattach aic_pcmcia_ca = {
	sizeof(struct aic_softc), aic_pcmcia_match, aic_isa_attach,
	aic_pcmcia_detach
};

STATIC int
aic_pcmcia_isa_attach(parent, match, aux, pc_link)
	struct device *parent;
	void *match;
	void *aux;
	struct pcmcia_link *pc_link;
{
	struct aic_softc *sc = match;
	struct isa_attach_args *ia = aux;

#ifdef AIC_DEBUG
	printf("aic_pcmcia_isa_attach: iobase %x\n", ia->ia_iobase);
	aic_debug |= AIC_SHOWSTART;
#endif

	sc->sc_iobase = ia->ia_iobase;
	if (aic_find(sc) != 0)
		return 0;

	ia->ia_msize = 0;
	ia->ia_iosize = 0x20;

#if 0
	sc->pcmcia_flags
	    = (pc_link->flags & PCMCIA_REATTACH) ? AIC_REATTACH : 0;
#endif
	return 1;
}

/* modify config entry */
STATIC int
aic_pcmcia_mod(pc_link, self, pc_cf, cf)
	struct pcmcia_link *pc_link;
	struct device *self;
	struct pcmcia_conf *pc_cf;
	struct cfdata *cf;
{
	int error;

	error = PCMCIA_BUS_CONFIG(pc_link->adapter, pc_link, self, pc_cf, cf);
	if (error) {
#ifdef DIAGNOSTIC
		printf("aic_pcmcia_mod: bus_config failed %d\n", error);
#endif
		return error;
	}

#ifdef PCMCIA_DEBUG
	pcmciadumpcf (pc_cf);
#endif

#if 1
	if (pc_cf->cfgtype == 0)
		pc_cf->cfgtype = CFGENTRYID;
	/* XXX; force to set PCMCIA_IO0_CS16, not PCMCIA_IO0_16BIT */
	pc_cf->io[0].flags |= PCMCIA_MAP_8 | PCMCIA_MAP_16;
#else
	pc_cf->cfgtype = DOSRESET | 9;	/* 9: 0x340, 8: 0x140 */
#endif

	return 0;
}

STATIC int
aic_remove(pc_link, self)
	struct pcmcia_link *pc_link;
	struct device *self;
{
#if 0
	struct aic_softc *sc = (void *) self;

	sc->pcmcia_flags = AIC_ABSENT;
	isa_intr_disestablish(sc->sc_bc, sc->sc_ih);
#endif
	return PCMCIA_BUS_UNCONFIG(pc_link->adapter, pc_link);
}

STATIC struct pcmcia_aic {
	struct pcmcia_device pcd;
} pcmcia_aic = {
	{ "PCMCIA Slim SCSI", aic_pcmcia_mod, aic_pcmcia_isa_attach, NULL,
	  aic_remove },
};

struct pcmciadevs pcmcia_aic_devs[] = {
	{
		"aic", 0, "Adaptec, Inc.", "APA-1460 SCSI Host Adapter",
		"Version 0.01", NULL,
		(void *) -1, (void *) &pcmcia_aic
	},
};
#define naic_pcmcia_devs sizeof(pcmcia_aic_devs)/sizeof(pcmcia_aic_devs[0])

int
aic_pcmcia_match(parent, match, aux)
	struct device *parent;
	void *match, *aux;
{

	return pcmcia_slave_match(parent, match, aux, pcmcia_aic_devs,
				  naic_pcmcia_devs);
}

#if 0
void
aic_pcmcia_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct aic_softc *sc = (void *)self;
	struct isa_attach_args *ia = aux;
	bus_chipset_tag_t bc = ia->ia_bc;
	bus_io_handle_t ioh;

#ifdef AIC_DEBUG
	printf("aic_pcmcia_attach\n");
	aic_debug |= AIC_SHOWSTART;
#endif
	if (bus_io_map(bc, ia->ia_iobase, ia->ia_iosize, &ioh)) {
#ifdef DIAGNOSTIC
		printf("aic_pcmcia_attach: can't map I/O space");
#endif
		return;
	}
	printf(" APA-1460");
	aicattach(parent, self, aux);
}
#endif

int
aic_pcmcia_detach(self)
	struct device *self;
{

	/*
	 * XXX Currently, no detach.
	 */
	return EBUSY;
}
