/*	$NetBSD$	*/
/*
 * Copyright (c) 1996, John T. Kohl.  All rights reserved.
 * Copyright (c) 1995, Stefan Grefen.
 * Copyright (c) 1994, 1995 Charles M. Hannum.  All rights reserved.
 *
 * Copyright (C) 1993, David Greenman.  This software may be used, modified,
 * copied, distributed, and sold, in both source and binary form provided that
 * the above copyright and these terms are retained.  Under no circumstances is
 * the author responsible for the proper functioning of this software, nor does
 * the author assume any responsibility for damages incurred with its use.
 *
 */

#include "bpfilter.h"

#include <sys/param.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <sys/syslog.h>
#include <sys/select.h>
#include <sys/device.h>

#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_types.h>
#include <net/netisr.h>

#ifdef INET
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/in_var.h>
#include <netinet/ip.h>
#include <netinet/if_ether.h>
#endif

#ifdef NS
#include <netns/ns.h>
#include <netns/ns_if.h>
#endif

#if NBPFILTER > 0
#include <net/bpf.h>
#include <net/bpfdesc.h>
#endif

#include <machine/cpu.h>
#include <machine/pio.h>
#include <machine/bus.h>

#include <dev/isa/isareg.h>
#include <dev/isa/isavar.h>

#include <dev/ic/dp8390reg.h>
#include <dev/ic/dp8390var.h>

#include <dev/pcmcia/pcmciavar.h>

int ed_pcmcia_match __P((struct device *, void *, void *));
void ed_pcmcia_attach __P((struct device *, struct device *, void *));
int ed_pcmcia_detach __P((struct device *));

struct cfattach ed_pcmcia_ca = {
	sizeof(struct ed_softc), ed_pcmcia_match, ed_pcmcia_attach,
	ed_pcmcia_detach
};

static int ed_pcmcia_isa_attach __P((struct device *, void *,
				     void *, struct pcmcia_link *));

static int edmod __P((struct pcmcia_link *, struct device *,
		      struct pcmcia_conf *, struct cfdata *cf));

static int ed_remove __P((struct pcmcia_link *, struct device *));

/* additional setup needed for pcmcia devices */
static int
ed_pcmcia_isa_attach(parent, match, aux, pc_link)
	struct device *parent;
	void *match;
	void *aux;
	struct pcmcia_link *pc_link;
{
	struct ed_softc *sc = match;
	struct cfdata *cf = sc->sc_dev.dv_cfdata;
	struct isa_attach_args *ia = aux;
	struct pcmciadevs *dev=pc_link->device;
	int err;
	extern int ifqmaxlen;
	u_char enaddr[ETHER_ADDR_LEN];

	if ((int)dev->param != -1)
		err = pcmcia_read_cis(pc_link, enaddr, 
				      (int) dev->param, ETHER_ADDR_LEN);
	else
		err = 0;
	if (err)
		printf("%s: attaching ed: cannot read cis info %d\n",
		       parent->dv_xname, err);

	if (ed_find_Novell(sc, cf, ia)) {
		delay(100);
		if ((int)dev->param != -1) {
		    err = pcmcia_read_cis(pc_link, sc->sc_arpcom.ac_enaddr,
				      (int) dev->param, ETHER_ADDR_LEN);
		    if (err) {
			    printf("Cannot read cis info %d\n", err);
			    return 0;
		    }
		    if(bcmp(enaddr, sc->sc_arpcom.ac_enaddr, ETHER_ADDR_LEN)) {
			    printf("ENADDR MISMATCH %s ",
				   ether_sprintf(sc->sc_arpcom.ac_enaddr));
			    printf("- %s\n", ether_sprintf(enaddr));
			    bcopy(enaddr,sc->sc_arpcom.ac_enaddr,
					    ETHER_ADDR_LEN);
		    }
		}
		/* clear ED_NOTPRESENT, set ED_REATTACH if needed */
		sc->spec_flags=pc_link->flags&PCMCIA_REATTACH?ED_REATTACH:0;
		sc->type_str = dev->model;
		sc->sc_arpcom.ac_if.if_snd.ifq_maxlen=ifqmaxlen;
		return 1;
	} else
	    return 0;
}

/* modify config entry */
static int
edmod(pc_link, self, pc_cf, cf) 
	struct pcmcia_link *pc_link;
	struct device *self;
	struct pcmcia_conf *pc_cf;
	struct cfdata *cf;
{
	int err;
/*	struct pcmciadevs *dev=pc_link->device;*/
/*	struct ed_softc *sc = (void *)self;*/
	int svec_card =  pc_cf->memwin  == 5;
	int de650_0 = (pc_cf->memwin != 0) && !svec_card;
	err = PCMCIA_BUS_CONFIG(pc_link->adapter, pc_link, self, pc_cf, cf);
	if (err)
		return err;

	if (svec_card) {
		pc_cf->memwin = 0;
#if 0
		pc_cf->cfgid = 32;  /* Try this if it still doesn't work */
		pc_cf->cfgid |= 32;  /* or Try this if it still doesn't work */
#endif
	}
	if (de650_0) {
		pc_cf->io[0].flags =
		    (pc_cf->io[0].flags&~PCMCIA_MAP_16)|PCMCIA_MAP_8;
		pc_cf->memwin = 0;
		pc_cf->cfgtype = DOSRESET|1;
	}
	else {
		/* still wrong in CIS; fix it here */
		pc_cf->io[0].flags = PCMCIA_MAP_8|PCMCIA_MAP_16;
		pc_cf->cfgtype = 1;
	}

	return err;
}


static int
ed_remove(pc_link,self) 
	struct pcmcia_link *pc_link;
	struct device *self;
{
	struct ed_softc *sc = (void *)self;
	struct ifnet *ifp = &sc->sc_arpcom.ac_if;
	if_down(ifp);
	edstop(sc);
	shutdownhook_disestablish(sc->sc_sh);
	ifp->if_flags &= ~(IFF_RUNNING|IFF_UP);
	sc->spec_flags |= ED_NOTPRESENT;
	isa_intr_disestablish(sc->sc_bc, sc->sc_ih);
	return PCMCIA_BUS_UNCONFIG(pc_link->adapter, pc_link);
}

static struct pcmcia_dlink {
	struct pcmcia_device pcd;
} pcmcia_dlink = {
	{"PCMCIA Novell compatible", edmod, ed_pcmcia_isa_attach,
	 NULL, ed_remove}
};

struct pcmciadevs pcmcia_ed_devs[]={
      { "ed", 0, "D-Link", "DE-650", "Ver 01.00", NULL, (void *) -1,
	(void *)&pcmcia_dlink },
      { "ed", 0, "D-Link", "DE-650", "", NULL, (void *) 0x40,
	(void *)&pcmcia_dlink },
      { "ed", 0, "LINKSYS", "E-CARD", "Ver 01.00", NULL, (void *)-1,
        (void *)&pcmcia_dlink },
      { "ed", 0, "IBM Corp.", "Ethernet", "0933495", NULL, (void *) 0xff0,
	(void *)&pcmcia_dlink },
      { "ed", 0, "Socket Communications Inc",
	"Socket EA PCMCIA LAN Adapter Revision D", "Ethernet ID 000000000000",
	NULL, (void *) -1,
	(void *)&pcmcia_dlink },
      /* something screwed up in ports requested */
      { "ed", 0, "SVEC", "FD605 PCMCIA EtherNet Card", "V1-1", NULL,
	(void *)-1, (void *)&pcmcia_dlink },
      { "ed", 0, "PCMCIA", "UE2212", "ETHERNET", "R01", (void *)-1,
        (void *)&pcmcia_dlink },
#if 0
      /* not quite right for ethernet adress */
      { "ed", 0, "PMX   ", "PE-200", "ETHERNET", "R01", (void *)-1,
        (void *)&pcmcia_dlink },
#endif
};

#define ned_pcmcia_devs sizeof(pcmcia_ed_devs)/sizeof(pcmcia_ed_devs[0])

int
ed_pcmcia_match(parent, match, aux)
	struct device *parent;
	void *match, *aux;
{
	return pcmcia_slave_match(parent, match, aux, pcmcia_ed_devs,
				  ned_pcmcia_devs);
}

void
ed_pcmcia_attach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct ed_softc *sc = (void *)self;
	struct isa_attach_args *ia = aux;

#if 0
	bus_chipset_tag_t bc;
	bus_io_handle_t ioh;
	/*
	 * XXX Should re-map io and mem, but can't
	 * XXX until we squish "indirect" brokenness.
	 */
	bc = sc->sc_bc;			/* XXX */
	ioh = sc->sc_ioh;		/* XXX */
#endif

	sc->sc_delayioh = ia->ia_delayioh;

	edattach(sc, sc->sc_bc, sc->sc_ioh);

	sc->sc_ih = isa_intr_establish(ia->ia_ic, ia->ia_irq, IST_EDGE,
				       IPL_NET, edintr, sc);

}

/*
 * No detach; network devices are too well linked into the rest of the
 * kernel.
 */
int
ed_pcmcia_detach(self)
	struct device *self;
{
	return EBUSY;
}

